import numpy as np
from cpymad.madx import Madx
import xtrack as xt

mad = Madx()


mad.call('../../test_data/hllhc15_thick/lhc.seq')
mad.call('../../test_data/hllhc15_thick/hllhc_sequence.madx')
mad.call('../../test_data/hllhc15_thick/opt_round_150_1500.madx')

# mad.call('../../test_data/hllhc14_no_errors_with_coupling_knobs/lhcb1_seq.madx')

mad.input('beam, sequence=lhcb1, particle=proton, pc=7000;')
mad.use(sequence='lhcb1')
mad.globals['on_x5'] = 300
tw_chr = mad.twiss(chrom=True)
twm = xt.Table(tw_chr)

line = xt.Line.from_madx_sequence(mad.sequence.lhcb1, deferred_expressions=True)

# tt = line.get_table()
# tt_sol = tt.rows[tt.element_type == 'Solenoid']
# for nn in tt_sol.name:
#     line.element_dict[nn] = xt.Drift(length=line[nn].length)

line.particle_ref = xt.Particles(mass0=mad.sequence.lhcb1.beam.mass*1e9,
                                    q0=mad.sequence.lhcb1.beam.charge,
                                    gamma0=mad.sequence.lhcb1.beam.gamma)

tw = line.twiss(method='4d')
nlchr = line.get_non_linear_chromaticity(delta0_range=(-1e-4, 1e-4),
                                        num_delta=21, fit_order=1)
tw_fw = line.twiss(start='ip4', end='ip6', init_at='ip4',
              x=tw['x', 'ip4'], px=tw['px', 'ip4'],
              y=tw['y', 'ip4'], py=tw['py', 'ip4'],
              betx=tw['betx', 'ip4'], bety=tw['bety', 'ip4'],
              alfx=tw['alfx', 'ip4'], alfy=tw['alfy', 'ip4'],
              dx=tw['dx', 'ip4'], dpx=tw['dpx', 'ip4'],
              dy=tw['dy', 'ip4'], dpy=tw['dpy', 'ip4'],
              ddx=tw['ddx', 'ip4'], ddy=tw['ddy', 'ip4'],
              ddpx=tw['ddpx', 'ip4'], ddpy=tw['ddpy', 'ip4'],
              compute_chromatic_properties=True)

tw_bw = line.twiss(start='ip4', end='ip6', init_at='ip6',
              x=tw['x', 'ip6'], px=tw['px', 'ip6'],
              y=tw['y', 'ip6'], py=tw['py', 'ip6'],
              betx=tw['betx', 'ip6'], bety=tw['bety', 'ip6'],
              alfx=tw['alfx', 'ip6'], alfy=tw['alfy', 'ip6'],
              dx=tw['dx', 'ip6'], dpx=tw['dpx', 'ip6'],
              dy=tw['dy', 'ip6'], dpy=tw['dpy', 'ip6'],
              ddx=tw['ddx', 'ip6'], ddy=tw['ddy', 'ip6'],
              ddpx=tw['ddpx', 'ip6'], ddpy=tw['ddpy', 'ip6'],
              compute_chromatic_properties=True)


tw_fw_mad = line.twiss(start='ip4', end='ip6', init_at='ip4',
              x=twm['x', 'ip4:1'], px=twm['px', 'ip4:1'],
              y=twm['y', 'ip4:1'], py=twm['py', 'ip4:1'],
              betx=twm['betx', 'ip4:1'], bety=twm['bety', 'ip4:1'],
              alfx=twm['alfx', 'ip4:1'], alfy=twm['alfy', 'ip4:1'],
              dx=twm['dx', 'ip4:1'], dpx=twm['dpx', 'ip4:1'],
              dy=twm['dy', 'ip4:1'], dpy=twm['dpy', 'ip4:1'],
              ddx=twm['ddx', 'ip4:1']*2, ddy=twm['ddy', 'ip4:1']*2,
              ddpx=twm['ddpx', 'ip4:1']*2, ddpy=twm['ddpy', 'ip4:1']*2,
              compute_chromatic_properties=True)

ddy_mad = np.interp(tw_fw.s, twm.s, twm.ddy)
ddx_mad = np.interp(tw_fw.s, twm.s, twm.ddx)

nlchr = line.get_non_linear_chromaticity(delta0_range=(-1e-4, 1e-4),
                                        num_delta=21, fit_order=2)

tw_mad = []
for dd in nlchr.delta0:
    mad.input(f'twiss, deltap={dd};')
    df = mad.table.twiss.dframe()
    tw_mad.append(xt.Table({kk: df[kk].values for kk in df.columns}))

location = 'ip3'

x_mad = np.array([tt['x', location + ':1'] for tt in tw_mad])
px_mad = np.array([tt['px', location + ':1'] for tt in tw_mad])
y_mad = np.array([tt['y', location + ':1'] for tt in tw_mad])
py_mad = np.array([tt['py', location + ':1'] for tt in tw_mad])

x_xs = np.array([tt['x', location] for tt in nlchr.twiss])
px_xs = np.array([tt['px', location] for tt in nlchr.twiss])
y_xs = np.array([tt['y', location] for tt in nlchr.twiss])
py_xs = np.array([tt['py', location] for tt in nlchr.twiss])
qx_xs = np.array([tt['qx'] for tt in nlchr.twiss])
qy_xs = np.array([tt['qy'] for tt in nlchr.twiss])
delta = np.array([tt['delta', location] for tt in nlchr.twiss])

pmad_x = np.polyfit(delta, x_mad, 3)
pmad_px = np.polyfit(delta, px_mad, 3)
pmad_y = np.polyfit(delta, y_mad, 3)
pmad_py = np.polyfit(delta, py_mad, 3)

pxs_x = np.polyfit(delta, x_xs, 3)
pxs_px = np.polyfit(delta, px_xs, 3)
pxs_y = np.polyfit(delta, y_xs, 3)
pxs_py = np.polyfit(delta, py_xs, 3)
pxs_qx = np.polyfit(delta, qx_xs, 3)
pxs_qy = np.polyfit(delta, qy_xs, 3)

assert np.allclose(delta, nlchr.delta0, atol=1e-6, rtol=0)
assert np.allclose(tw['dx', location], pxs_x[-2], atol=0, rtol=1e-4)
assert np.allclose(tw['dpx', location], pxs_px[-2], atol=0, rtol=1e-4)
assert np.allclose(tw['dy', location], pxs_y[-2], atol=0, rtol=1e-4)
assert np.allclose(tw['dpy', location], pxs_py[-2], atol=0, rtol=1e-4)
assert np.allclose(tw['ddx', location], 2*pxs_x[-3], atol=0, rtol=1e-4)
assert np.allclose(tw['ddpx', location], 2*pxs_px[-3], atol=0, rtol=1e-4)
assert np.allclose(tw['ddy', location], 2*pxs_y[-3], atol=0, rtol=1e-4)
assert np.allclose(tw['ddpy', location], 2*pxs_py[-3], atol=0, rtol=1e-4)
assert np.isclose(tw['dqx'], pxs_qx[-2], atol=0, rtol=1e-3)
assert np.isclose(tw['ddqx'], pxs_qx[-3]*2, atol=0, rtol=1e-4)
assert np.isclose(tw['dqy'], pxs_qy[-2], atol=0, rtol=1e-3)
assert np.isclose(tw['ddqy'], pxs_qy[-3]*2, atol=0, rtol=1e-4)

assert np.isclose(nlchr['dqx'], pxs_qx[-2], atol=0, rtol=2e-3)
assert np.isclose(nlchr['dqy'], pxs_qy[-2], atol=0, rtol=2e-3)
assert np.isclose(nlchr['ddqx'], pxs_qx[-3]*2, atol=0, rtol=1e-4)
assert np.isclose(nlchr['ddqy'], pxs_qy[-3]*2, atol=0, rtol=1e-4)

tw_part = tw.rows['ip4':'ip6']
assert np.allclose(tw_part['ddx'], tw_fw.rows[:-1]['ddx'], atol=1e-2, rtol=0)
assert np.allclose(tw_part['ddy'], tw_fw.rows[:-1]['ddy'], atol=1e-2, rtol=0)
assert np.allclose(tw_part['ddpx'], tw_fw.rows[:-1]['ddpx'], atol=1e-3, rtol=0)
assert np.allclose(tw_part['ddpy'], tw_fw.rows[:-1]['ddpy'], atol=1e-3, rtol=0)
assert np.allclose(tw_part['dx'], tw_bw.rows[:-1]['dx'], atol=1e-2, rtol=0)
assert np.allclose(tw_part['dy'], tw_bw.rows[:-1]['dy'], atol=1e-2, rtol=0)
assert np.allclose(tw_part['dpx'], tw_bw.rows[:-1]['dpx'], atol=1e-3, rtol=0)
assert np.allclose(tw_part['dpy'], tw_bw.rows[:-1]['dpy'], atol=1e-3, rtol=0)


import matplotlib.pyplot as plt
plt.figure(1)
ax1 = plt.subplot(4, 1, 1)
plt.plot(tw.s, tw.ddx)
plt.plot(tw_chr.s, 2 * tw_chr.ddx) # in MAD-X ddx = 0.5 d2x/ddelta2
plt.plot(tw_fw.s, tw_fw.ddx, '.')
plt.plot(tw_bw.s, tw_bw.ddx, '-')
plt.ylabel('ddx')
plt.subplot(4, 1, 2, sharex=ax1)
plt.plot(tw.s, tw.ddy)
plt.plot(tw_chr.s, tw_chr.ddy * 2)
plt.plot(tw_fw.s, tw_fw.ddy, '.')
plt.plot(tw_bw.s, tw_bw.ddy, '-')
plt.ylabel('ddy')
plt.subplot(4, 1, 3, sharex=ax1)
plt.plot(tw.s, tw.ddpx)
plt.plot(tw_chr.s, tw_chr.ddpx * 2)
plt.plot(tw_fw.s, tw_fw.ddpx, '.')
plt.plot(tw_bw.s, tw_bw.ddpx, '-')
plt.ylabel('ddpx')
plt.subplot(4, 1, 4, sharex=ax1)
plt.plot(tw.s, tw.ddpy)
plt.plot(tw_chr.s, tw_chr.ddpy * 2)
plt.plot(tw_fw.s, tw_fw.ddpy, '.')
plt.plot(tw_bw.s, tw_bw.ddpy, '-')
plt.ylabel('ddpy')

plt.figure(2)
plt.subplot(2, 1, 1)
plt.plot(nlchr.delta0*1e3, x_xs, label='xsuite')
plt.plot(nlchr.delta0*1e3, x_mad, label='madx')
plt.ylabel('x')
plt.legend()
plt.subplot(2, 1, 2)
plt.plot(nlchr.delta0*1e3, y_xs, label='xsuite')
plt.plot(nlchr.delta0*1e3, y_mad, label='madx')
plt.ylabel('y')
plt.legend()
plt.xlabel(r'$\delta$ [1e-3]')




plt.show()


import json

import numpy as np
import xtrack as xt
import xobjects as xo

with open('../../test_data/hllhc14_no_errors_with_coupling_knobs/line_b1.json',
            'r') as fid:
    dct = json.load(fid)
line = xt.Line.from_dict(dct)

line.build_tracker()

tw_before = line.twiss()

GreaterThan = xt.GreaterThan
LessThan = xt.LessThan

tw0 = line.twiss()
opt = line.match(
    name='bump',
    solve=False,
    solver='jacobian',
    # Portion of the beam line to be modified and initial conditions
    start='mq.33l8.b1',
    end='mq.17l8.b1',
    init=tw0, init_at=xt.START,
    # Dipole corrector strengths to be varied
    vary=[
        xt.Vary(name='acbv32.l8b1', step=1e-10, weight=0.7),
        xt.Vary(name='acbv28.l8b1', step=1e-10, weight=0.3),
        xt.Vary(name='acbv26.l8b1', step=1e-10),
        xt.Vary(name='acbv24.l8b1', step=1e-10),
        xt.Vary(name='acbv22.l8b1', step=1e-10, limits=[-38e-6, 38e-6], weight=1000),
        xt.Vary(name='acbv18.l8b1', step=1e-10),
    ],
    targets=[
        xt.Target('py', at='mb.b26l8.b1', value=0, tol=1e-6, weight=1e3),
        xt.Target('y', at='mb.b26l8.b1', value=3e-3, tol=1e-4),
        xt.Target('y', at='mq.17l8.b1', value=tw0, tol=1e-6),
        xt.Target('py', at='mq.17l8.b1', value=tw0, tol=1e-7, weight=1e3),
    ]
)

# Check target_mismatch
assert opt.name == 'bump'
ts = opt.target_status(ret=True)
assert len(ts) == 4
assert np.all(ts.tol_met == np.array([True, False, True, True]))
tm = opt.target_mismatch(ret=True)
assert len(tm) == 1
assert tm.id[0] == 1

opt.solve()

# I want to limit the negative excursion ot the bump
opt2 = opt.clone(name='limit',
    add_targets=[
        xt.Target('y', GreaterThan(-2e-3), at='mq.30l8.b1', tol=1e-6),
        xt.Target('y', GreaterThan(-1e-3), at='mq.30l8.b1', tol=1e-6)])
opt2.solve()

assert opt2.name == 'limit'
assert len(opt2.targets) == 6
tm = opt2.target_mismatch(ret=True)
assert(len(tm) == 0)

tw = line.twiss()

assert np.isclose(tw['y', 'mq.33l8.b1'], 0, atol=1e-6, rtol=0)
assert np.isclose(tw['y', 'mq.17l8.b1'], 0, atol=1e-6, rtol=0)
assert np.isclose(tw['py', 'mq.17l8.b1'], 0, atol=1e-8, rtol=0)
assert np.isclose(tw['py', 'mq.33l8.b1'], 0, atol=1e-6, rtol=0)

assert np.isclose(tw['y', 'mb.b26l8.b1'], 3e-3, atol=1e-6, rtol=0)
assert np.isclose(tw['py', 'mb.b26l8.b1'], 0, atol=1e-8, rtol=0)

assert np.isclose(tw['y', 'mq.30l8.b1'], -1e-3, atol=1e-6, rtol=0)
assert np.isclose(line.vars['acbv22.l8b1']._value, 38e-6, atol=0, rtol=0.02)

# Test variable in inequality
line['myvar'] = -5e-3
opt3 = opt2.clone(name='ineq',
    add_targets=[
        xt.Target('y', GreaterThan(line.ref['myvar']), at='mq.30l8.b1', tol=1e-6)])

assert len(opt3.target_mismatch(ret=True)) == 0
assert opt3.target_status(ret=True).residue[-1] == 0

line['myvar'] = -0.5e-3
assert len(opt3.target_mismatch(ret=True)) == 1
xo.assert_allclose(opt3.target_status(ret=True).residue[-1], -0.5e-3,
                   atol=1e-5, rtol=0)

opt3.solve()
assert len(opt3.target_mismatch(ret=True)) == 0



import time

import xtrack as xt

import xtrack._temp.lhc_match as lm

default_tol = {None: 1e-8, 'betx': 1e-6, 'bety': 1e-6} # to have no rematching w.r.t. madx

collider = xt.Environment.from_json(
    "../../test_data/hllhc15_thick/hllhc15_collider_thick.json")
collider.build_trackers()
collider.vars.load_madx_optics_file(
    "../../test_data/hllhc15_thick/opt_round_150_1500.madx")

lm.set_var_limits_and_steps(collider)

tw0 = collider.twiss()

d_mux_15_b1 = 0.07
d_muy_15_b1 = 0.05
d_mux_15_b2 = -0.1
d_muy_15_b2 = -0.12

staged_match = True

optimizers = {}

print('Matching ip15 phase:')
t1 = time.time()
opt = lm.change_phase_non_ats_arcs(collider,
    d_mux_15_b1=d_mux_15_b1, d_muy_15_b1=d_muy_15_b1,
    d_mux_15_b2=d_mux_15_b2, d_muy_15_b2=d_muy_15_b2,
    solve=True, default_tol=default_tol)
t2 = time.time()
optimizers['phase_15'] = opt

arc_periodic_solution = lm.get_arc_periodic_solution(collider)

optimizers.update({'b1':{}, 'b2':{}})

t_match_irs = {}
for bn in ['b1', 'b2']:

    tt_1 = time.time()
    line_name = f'lhc{bn}'

    muxip1_l = collider.varval[f'muxip1{bn}_l']
    muyip1_l = collider.varval[f'muyip1{bn}_l']
    muxip1_r = collider.varval[f'muxip1{bn}_r']
    muyip1_r = collider.varval[f'muyip1{bn}_r']

    muxip5_l = collider.varval[f'muxip5{bn}_l']
    muyip5_l = collider.varval[f'muyip5{bn}_l']
    muxip5_r = collider.varval[f'muxip5{bn}_r']
    muyip5_r = collider.varval[f'muyip5{bn}_r']

    muxip2 = collider.varval[f'muxip2{bn}']
    muyip2 = collider.varval[f'muyip2{bn}']
    muxip4 = collider.varval[f'muxip4{bn}']
    muyip4 = collider.varval[f'muyip4{bn}']
    muxip6 = collider.varval[f'muxip6{bn}']
    muyip6 = collider.varval[f'muyip6{bn}']
    muxip8 = collider.varval[f'muxip8{bn}']
    muyip8 = collider.varval[f'muyip8{bn}']

    mux12 = collider.varval[f'mux12{bn}']
    muy12 = collider.varval[f'muy12{bn}']
    mux45 = collider.varval[f'mux45{bn}']
    muy45 = collider.varval[f'muy45{bn}']
    mux56 = collider.varval[f'mux56{bn}']
    muy56 = collider.varval[f'muy56{bn}']
    mux81 = collider.varval[f'mux81{bn}']
    muy81 = collider.varval[f'muy81{bn}']

    betx_ip1 = collider.varval[f'betxip1{bn}']
    bety_ip1 = collider.varval[f'betyip1{bn}']
    betx_ip5 = collider.varval[f'betxip5{bn}']
    bety_ip5 = collider.varval[f'betyip5{bn}']

    betx_ip2 = collider.varval[f'betxip2{bn}']
    bety_ip2 = collider.varval[f'betyip2{bn}']

    alfx_ip3 = collider.varval[f'alfxip3{bn}']
    alfy_ip3 = collider.varval[f'alfyip3{bn}']
    betx_ip3 = collider.varval[f'betxip3{bn}']
    bety_ip3 = collider.varval[f'betyip3{bn}']
    dx_ip3 = collider.varval[f'dxip3{bn}']
    dpx_ip3 = collider.varval[f'dpxip3{bn}']
    mux_ir3 = collider.varval[f'muxip3{bn}']
    muy_ir3 = collider.varval[f'muyip3{bn}']

    alfx_ip4 = collider.varval[f'alfxip4{bn}']
    alfy_ip4 = collider.varval[f'alfyip4{bn}']
    betx_ip4 = collider.varval[f'betxip4{bn}']
    bety_ip4 = collider.varval[f'betyip4{bn}']
    dx_ip4 = collider.varval[f'dxip4{bn}']
    dpx_ip4 = collider.varval[f'dpxip4{bn}']

    alfx_ip6 = collider.varval[f'alfxip6{bn}']
    alfy_ip6 = collider.varval[f'alfyip6{bn}']
    betx_ip6 = collider.varval[f'betxip6{bn}']
    bety_ip6 = collider.varval[f'betyip6{bn}']
    dx_ip6 = collider.varval[f'dxip6{bn}']
    dpx_ip6 = collider.varval[f'dpxip6{bn}']

    alfx_ip7 = collider.varval[f'alfxip7{bn}']
    alfy_ip7 = collider.varval[f'alfyip7{bn}']
    betx_ip7 = collider.varval[f'betxip7{bn}']
    bety_ip7 = collider.varval[f'betyip7{bn}']
    dx_ip7 = collider.varval[f'dxip7{bn}']
    dpx_ip7 = collider.varval[f'dpxip7{bn}']
    mux_ir7 = collider.varval[f'muxip7{bn}']
    muy_ir7 = collider.varval[f'muyip7{bn}']

    alfx_ip8 = collider.varval[f'alfxip8{bn}']
    alfy_ip8 = collider.varval[f'alfyip8{bn}']
    betx_ip8 = collider.varval[f'betxip8{bn}']
    bety_ip8 = collider.varval[f'betyip8{bn}']
    dx_ip8 = collider.varval[f'dxip8{bn}']
    dpx_ip8 = collider.varval[f'dpxip8{bn}']

    tw_sq_a81_ip1_a12 = lm.propagate_optics_from_beta_star(collider, ip_name='ip1',
            line_name=f'lhc{bn}', start=f's.ds.r8.{bn}', end=f'e.ds.l2.{bn}',
            beta_star_x=betx_ip1, beta_star_y=bety_ip1)

    tw_sq_a45_ip5_a56 = lm.propagate_optics_from_beta_star(collider, ip_name='ip5',
            line_name=f'lhc{bn}', start=f's.ds.r4.{bn}', end=f'e.ds.l6.{bn}',
            beta_star_x=betx_ip5, beta_star_y=bety_ip5)

    (mux_ir2_target, muy_ir2_target, mux_ir4_target, muy_ir4_target,
     mux_ir6_target, muy_ir6_target, mux_ir8_target, muy_ir8_target
        ) = lm.compute_ats_phase_advances_for_auxiliary_irs(line_name,
            tw_sq_a81_ip1_a12, tw_sq_a45_ip5_a56,
            muxip1_l, muyip1_l, muxip1_r, muyip1_r,
            muxip5_l, muyip5_l, muxip5_r, muyip5_r,
            muxip2, muyip2, muxip4, muyip4, muxip6, muyip6, muxip8, muyip8,
            mux12, muy12, mux45, muy45, mux56, muy56, mux81, muy81)

    print(f"Matching IR2 {bn}")
    opt = lm.rematch_ir2(collider, line_name=f'lhc{bn}',
                boundary_conditions_left=tw_sq_a81_ip1_a12,
                boundary_conditions_right=arc_periodic_solution[f'lhc{bn}']['23'],
                mux_ir2=mux_ir2_target, muy_ir2=muy_ir2_target,
                betx_ip2=betx_ip2, bety_ip2=bety_ip2,
                solve=True, staged_match=staged_match,
                default_tol=default_tol)
    optimizers[bn]['ir2'] = opt

    print(f"Matching IR3 {bn}")
    opt = lm.rematch_ir3(collider=collider, line_name=f'lhc{bn}',
                boundary_conditions_left=arc_periodic_solution[f'lhc{bn}']['23'],
                boundary_conditions_right=arc_periodic_solution[f'lhc{bn}']['34'],
                mux_ir3=mux_ir3, muy_ir3=muy_ir3,
                alfx_ip3=alfx_ip3, alfy_ip3=alfy_ip3,
                betx_ip3=betx_ip3, bety_ip3=bety_ip3,
                dx_ip3=dx_ip3, dpx_ip3=dpx_ip3,
                solve=True, staged_match=staged_match, default_tol=default_tol)
    optimizers[bn]['ir3'] = opt

    print(f"Matching IR4 {bn}")
    opt = lm.rematch_ir4(collider=collider, line_name=f'lhc{bn}',
                boundary_conditions_left=arc_periodic_solution[f'lhc{bn}']['34'],
                boundary_conditions_right=tw_sq_a45_ip5_a56,
                mux_ir4=mux_ir4_target, muy_ir4=muy_ir4_target,
                alfx_ip4=alfx_ip4, alfy_ip4=alfy_ip4,
                betx_ip4=betx_ip4, bety_ip4=bety_ip4,
                dx_ip4=dx_ip4, dpx_ip4=dpx_ip4,
                solve=True, staged_match=staged_match, default_tol=default_tol)
    optimizers[bn]['ir4'] = opt

    print(f"Matching IP6 {bn}")
    opt = lm.rematch_ir6(collider=collider, line_name=f'lhc{bn}',
                boundary_conditions_left=tw_sq_a45_ip5_a56,
                boundary_conditions_right=arc_periodic_solution[f'lhc{bn}']['67'],
                mux_ir6=mux_ir6_target, muy_ir6=muy_ir6_target,
                alfx_ip6=alfx_ip6, alfy_ip6=alfy_ip6,
                betx_ip6=betx_ip6, bety_ip6=bety_ip6,
                dx_ip6=dx_ip6, dpx_ip6=dpx_ip6,
                solve=True, staged_match=staged_match, default_tol=default_tol)
    optimizers[bn]['ir6'] = opt

    print(f"Matching IP7 {bn}")
    opt = lm.rematch_ir7(collider=collider, line_name=f'lhc{bn}',
            boundary_conditions_left=arc_periodic_solution[f'lhc{bn}']['67'],
            boundary_conditions_right=arc_periodic_solution[f'lhc{bn}']['78'],
            mux_ir7=mux_ir7, muy_ir7=muy_ir7,
            alfx_ip7=alfx_ip7, alfy_ip7=alfy_ip7,
            betx_ip7=betx_ip7, bety_ip7=bety_ip7,
            dx_ip7=dx_ip7, dpx_ip7=dpx_ip7,
            solve=True, staged_match=staged_match, default_tol=default_tol)
    optimizers[bn]['ir7'] = opt

    print(f"Matching IP8 {bn}")
    opt = lm.rematch_ir8(collider=collider, line_name=f'lhc{bn}',
            boundary_conditions_left=arc_periodic_solution[f'lhc{bn}']['78'],
            boundary_conditions_right=tw_sq_a81_ip1_a12,
            mux_ir8=mux_ir8_target, muy_ir8=muy_ir8_target,
            alfx_ip8=alfx_ip8, alfy_ip8=alfy_ip8,
            betx_ip8=betx_ip8, bety_ip8=bety_ip8,
            dx_ip8=dx_ip8, dpx_ip8=dpx_ip8,
            solve=True, staged_match=staged_match, default_tol=default_tol)
    optimizers[bn]['ir8'] = opt

    tt_2 = time.time()
    t_match_irs[bn] = tt_2 - tt_1

t3 = time.time()
opt = lm.match_orbit_knobs_ip2_ip8(collider)
t4 = time.time()
optimizers['orbit_knobs'] = opt

tw = collider.twiss()

# Tunes
print('Tunes:')
print(f"  b1: qx={tw.lhcb1.qx:6f} qy={tw.lhcb1.qy:6f}")
print(f"  b2: qx={tw.lhcb2.qx:6f} qy={tw.lhcb2.qy:6f}")

print('IP15 phase before:')
print(f"  b1: d_mux={tw0.lhcb1['mux', 'ip5'] - tw0.lhcb1['mux', 'ip1']:6f} "
      f"      d_muy={tw0.lhcb1['muy', 'ip5'] - tw0.lhcb1['muy', 'ip1']:6f} ")
print(f"  b2: d_mux={tw0.lhcb2['mux', 'ip5'] - tw0.lhcb2['mux', 'ip1']:6f} "
      f"      d_muy={tw0.lhcb2['muy', 'ip5'] - tw0.lhcb2['muy', 'ip1']:6f} ")

print('IP15 phase after:')
print(f"  b1: d_mux={tw.lhcb1['mux', 'ip5'] - tw.lhcb1['mux', 'ip1']:6f} "
      f"      d_muy={tw.lhcb1['muy', 'ip5'] - tw.lhcb1['muy', 'ip1']:6f} ")
print(f"  b2: d_mux={tw.lhcb2['mux', 'ip5'] - tw.lhcb2['mux', 'ip1']:6f} "
      f"      d_muy={tw.lhcb2['muy', 'ip5'] - tw.lhcb2['muy', 'ip1']:6f} ")

print('IP15 phase shifts:')
print(f"  b1: d_mux={tw.lhcb1['mux', 'ip5'] - tw0.lhcb1['mux', 'ip5']:6f} "
            f"d_muy={tw.lhcb1['muy', 'ip5'] - tw0.lhcb1['muy', 'ip5']:6f} ")
print(f"  b2: d_mux={tw.lhcb2['mux', 'ip5'] - tw0.lhcb2['mux', 'ip5']:6f} "
            f"d_muy={tw.lhcb2['muy', 'ip5'] - tw0.lhcb2['muy', 'ip5']:6f} ")

print(f'Time match arcs: {t2-t1} s')
print(f'Time match IRs b1: {t_match_irs["b1"]}')
print(f'Time match IRs b2: {t_match_irs["b2"]}')
print(f'Time match orbit knobs: {t4-t3} s')

# # Open twisses to debug
# tw_b1_check = collider.lhcb1.twiss(
#     start='ip1', end='ip1.l1',
#     init=xt.TwissInit(betx=0.15, bety=0.15, element_name='ip1',
#                             line=collider.lhcb1))

# tw_b2_check = collider.lhcb2.twiss(
#     start='ip1', end='ip1.l1',
#     init=xt.TwissInit(betx=0.15, bety=0.15, element_name='ip1',
#                             line=collider.lhcb2))

# Save xsuite model
collider.to_json('collider_01_changed_ip15_phase.json')

# Save madx optics file
lm.gen_madx_optics_file_auto(collider, 'opt_round_150_1500_xs.madx')


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np
import xtrack as xt

T_rev = 23e-6 # 23 us
Delta_p0c = 450e9 / 10 * T_rev

line = xt.Line.from_json('../../test_data/sps_w_spacecharge/line_no_spacecharge.json')

energy_increase = xt.ReferenceEnergyIncrease(Delta_p0c=Delta_p0c)
line.append_element(energy_increase, 'energy_increase')
line.build_tracker()

particles = xt.Particles(p0c=26e9, zeta=np.linspace(-1, 1, 40))

line.track(particles, num_turns=500, turn_by_turn_monitor=True)

#!end-doc-part

rec = line.record_last_track
import matplotlib.pyplot as plt
plt.close('all')
plt.figure(1)
for ii in range(rec.x.shape[0]):
    mask = rec.state[ii, :]>0
    plt.plot(rec.zeta[ii, mask], rec.delta[ii, mask])

plt.grid(linestyle='--')
plt.xlabel('z [m]')
plt.ylabel(r'$\Delta p / p_0$')

## Quick check for stable phase

from scipy.constants import c as clight

# Assume only first cavity is active
frequency = line.get_elements_of_type(xt.Cavity)[0][0].frequency
voltage = line.get_elements_of_type(xt.Cavity)[0][0].voltage
#Assuming proton and beta=1
stable_z = np.arcsin(Delta_p0c/voltage)/frequency/2/np.pi*clight

p_co = line.find_closed_orbit()

assert np.isclose(p_co.zeta, stable_z, atol=0, rtol=1e-2)

plt.axvline(x=stable_z, color='black', linestyle='--')

plt.show()



import pandas as pd
import numpy as np

from scipy.constants import c as clight

import xtrack as xt
import xdeps as xd


fname = '../../test_data/psb_chicane/Ramp_and_RF_functions.dat'

df = pd.read_csv(fname, sep='\t', skiprows=2,
    names=['t_s', 'E_kin_GeV', 'V1_MV', 'phi1_rad', 'V2_MV', 'phi2_rad'])
E_kin_GeV = df.E_kin_GeV.values
t_s = df.t_s.values

line = xt.Line.from_json('psb_04_with_chicane_corrected_thin.json')
line.build_tracker()

# Attach energy program
ep = xt.EnergyProgram(t_s=t_s, kinetic_energy0=E_kin_GeV*1e9)
line.energy_program = ep

# Frequency program
freq_rev = line.energy_program.get_frev_at_t_s(t_s)
line.functions['fun_freq_rev'] = xd.FunctionPieceWiseLinear(x=t_s, y=freq_rev)
line.vars['freq_rev'] = line.functions['fun_freq_rev'](line.vars['t_turn_s'])
line.vars['freq_h1'] = line.vars['freq_rev']
line.vars['freq_h2'] = 2 * line.vars['freq_rev']

# voltage programs

# Shift phases to have the beam centered around zero

V1_MV = df.V1_MV.values
V2_MV = df.V2_MV.values # to have opposite slopes at z=0
phi1_rad = df.phi1_rad.values - np.pi
phi2_rad = df.phi2_rad.values - np.pi

line.functions['fun_volt_mv_h1'] = xd.FunctionPieceWiseLinear(x=t_s, y=V1_MV)
line.functions['fun_volt_mv_h2'] = xd.FunctionPieceWiseLinear(x=t_s, y=V2_MV)
line.vars['volt_mv_h1'] = line.functions['fun_volt_mv_h1'](line.vars['t_turn_s'])
line.vars['volt_mv_h2'] = line.functions['fun_volt_mv_h2'](line.vars['t_turn_s'])

# phase programs
line.functions['fun_phi_rad_h1'] = xd.FunctionPieceWiseLinear(x=t_s, y=phi1_rad)
line.functions['fun_phi_rad_h2'] = xd.FunctionPieceWiseLinear(x=t_s, y=phi2_rad)
line.vars['phi_rad_h1'] = line.functions['fun_phi_rad_h1'](line.vars['t_turn_s'])
line.vars['phi_rad_h2'] = line.functions['fun_phi_rad_h2'](line.vars['t_turn_s'])

# Setup cavities
line.element_refs['br1.acwf5l1.1'].voltage = line.vars['volt_mv_h1'] * 1e6
line.element_refs['br1.acwf5l1.2'].voltage = line.vars['volt_mv_h2'] * 1e6

line.element_refs['br1.acwf5l1.1'].lag = line.vars['phi_rad_h1'] * 360 / 2 / np.pi
line.element_refs['br1.acwf5l1.2'].lag = line.vars['phi_rad_h2'] * 360 / 2 / np.pi

line.element_refs['br1.acwf5l1.1'].frequency = line.vars['freq_h1']
line.element_refs['br1.acwf5l1.2'].frequency = line.vars['freq_h2']


# tw6d = line.twiss(method='6d')

t_rev = []
beta0 = []
gamma0 = []
f_h1 = []
f_h2 = []
lag_h1 = []
lag_h2 = []
volt_h1 = []
volt_h2 = []
for ii in range(len(t_s)):
    print(f'Computing twiss at t_s = {t_s[ii]:.4} s    ', end='\r', flush=True)
    line.vars['t_turn_s'] = t_s[ii]
    tt = line.twiss(method='4d')
    t_rev.append(tt.T_rev0)
    beta0.append(tt.beta0)
    gamma0.append(tt.gamma0)
    f_h1.append(line['br1.acwf5l1.1'].frequency)
    f_h2.append(line['br1.acwf5l1.2'].frequency)
    lag_h1.append(line['br1.acwf5l1.1'].lag)
    lag_h2.append(line['br1.acwf5l1.2'].lag)
    volt_h1.append(line['br1.acwf5l1.1'].voltage)
    volt_h2.append(line['br1.acwf5l1.2'].voltage)


line.vars['t_turn_s'] = 0


# Test tracking
p_test = line.build_particles(x_norm=0, zeta=np.linspace(0, 100., 20))
assert np.isclose(p_test.energy0[0] - p_test.mass0,  E_kin_GeV[0] * 1e9,
                  atol=0, rtol=1e-10)

line.enable_time_dependent_vars = True
n_turn_test = 10000
monitor = xt.ParticlesMonitor(num_particles=len(p_test.zeta), start_at_turn=0,
                              stop_at_turn=n_turn_test)
for ii in range(n_turn_test):
    if ii % 10 == 0:
        print(f'Tracking turn {ii}/{n_turn_test}     ', end='\r', flush=True)
    line.track(p_test, turn_by_turn_monitor=monitor)


t_test = 40e-3
p0c_test = ep.get_p0c_at_t_s(t_test)
p_test.update_p0c_and_energy_deviations(p0c_test)
ekin_test = p_test.energy0[0] - p_test.mass0

t_turn = line.energy_program.get_t_s_at_turn(np.arange(n_turn_test))


import matplotlib.pyplot as plt
plt.close('all')

plt.figure(1)
sp_ekin = plt.subplot(3,1,1)
plt.plot(t_s, E_kin_GeV)
plt.plot(t_test, ekin_test*1e-9, 'o')
plt.ylabel(r'$E_{kin}$ [GeV]')

sp_dekin = plt.subplot(3,1,2, sharex=sp_ekin)
# GeV/sec
dekin = (E_kin_GeV[1:] - E_kin_GeV[:-1])/(t_s[1:] - t_s[:-1])
plt.plot(t_s[:-1], dekin)
plt.ylabel(r'd$E_{kin}$/dt [GeV/s]')

sp_beta = plt.subplot(3,1,3, sharex=sp_ekin)
plt.plot(t_turn, monitor.beta0.T)
plt.ylabel(r'$\beta$')
plt.xlabel('t [s]')

# plot rf parameters
plt.figure(2)
plt.subplot(3,1,1)
plt.plot(t_s, f_h1, label='h1')
plt.plot(t_s, f_h2, label='h2')
plt.ylabel('frequency [Hz]')
plt.legend()
plt.subplot(3,1,2)
plt.plot(t_s, lag_h1, label='h1')
plt.plot(t_s, lag_h2, label='h2')
plt.ylabel('lag [deg]')
plt.legend()

plt.subplot(3,1,3)
plt.plot(t_s, volt_h1, label='h1')
plt.plot(t_s, volt_h2, label='h2')
plt.ylabel('voltage [V]')
plt.legend()
plt.xlabel('t [s]')

plt.figure(3)
colors = plt.cm.jet(np.linspace(0,1, len(p_test.zeta)))
for ii in range(len(colors)):
    plt.plot(monitor.zeta[ii, :], monitor.delta[ii, :], color=colors[ii])
plt.xlim(-300, 300)
plt.ylim(-5e-3, 5e-3)

plt.show()



import numpy as np

import numpy as np

import xtrack as xt
import xdeps as xd

import matplotlib.pyplot as plt

line = xt.Line.from_json('psb_02_with_chicane_time_functions.json')
line.insert_element(element=xt.Marker(), name='mker_match', at_s=79.874)
line.build_tracker()

line.vars['on_chicane_k0'] = 0
line.vars['on_chicane_k2'] = 0
tw0 = line.twiss()
line.vars['on_chicane_k0'] = 1
line.vars['on_chicane_k2'] = 1

###############################
# Correct tunes and beta beat #
###############################

t_correct = np.linspace(0, 5.5e-3, 30)

# LInk vars
# line.vars['kbrqd14corr'] = line.vars['kbrqd3corr']

kbrqf_corr_list = []
kbrqd_corr_list = []
kbrqd3corr_list = []
kbrqd14corr_list = []
for ii, tt in enumerate(t_correct):
    print(f'Correct tune at t = {tt * 1e3:.2f} ms   \n')
    line.vars['t_turn_s'] = tt

    line.match(
        #verbose=True,
        vary=[
            xt.Vary('kbrqfcorr', step=1e-4),
            xt.Vary('kbrqdcorr', step=1e-4),
            xt.Vary('kbrqd3corr', step=1e-4),
            xt.Vary('kbrqd14corr', step=1e-4),
        ],
        targets = [
            xt.Target('qx', value=tw0.qx, tol=1e-5, scale=1),
            xt.Target('qy', value=tw0.qy, tol=1e-5, scale=1),
            xt.Target('bety', at='mker_match',
                      value=tw0['bety', 'mker_match'], tol=1e-4, scale=100),
            xt.Target('alfy', at='mker_match',
                       value=tw0['alfy', 'mker_match'], tol=1e-4, scale=100)
        ]
    )

    kbrqf_corr_list.append(line.vars['kbrqfcorr']._value)
    kbrqd_corr_list.append(line.vars['kbrqdcorr']._value)
    kbrqd3corr_list.append(line.vars['kbrqd3corr']._value)
    kbrqd14corr_list.append(line.vars['kbrqd14corr']._value)

line.functions['fun_kqf_corr'] = xd.FunctionPieceWiseLinear(
    x=t_correct, y=kbrqf_corr_list)
line.functions['fun_kqd_corr'] = xd.FunctionPieceWiseLinear(
    x=t_correct, y=kbrqd_corr_list)
line.functions['fun_qd3_corr'] = xd.FunctionPieceWiseLinear(
    x=t_correct, y=kbrqd3corr_list)
line.functions['fun_qd14_corr'] = xd.FunctionPieceWiseLinear(
    x=t_correct, y=kbrqd14corr_list)

line.vars['on_chicane_tune_corr'] = 1
line.vars['kbrqfcorr'] = (line.vars['on_chicane_tune_corr']
                            * line.functions.fun_kqf_corr(line.vars['t_turn_s']))
line.vars['kbrqdcorr'] = (line.vars['on_chicane_tune_corr']
                            * line.functions.fun_kqd_corr(line.vars['t_turn_s']))

line.vars['on_chicane_beta_corr'] = 1
line.vars['kbrqd3corr'] = (line.vars['on_chicane_beta_corr']
                         * line.functions.fun_qd3_corr(line.vars['t_turn_s']))
line.vars['kbrqd14corr'] = (line.vars['on_chicane_beta_corr']
                        * line.functions.fun_qd14_corr(line.vars['t_turn_s']))

# Save
line.vars['t_turn_s'] = 0
line.to_json('psb_03_with_chicane_corrected.json')


t_test = np.linspace(0, 6e-3, 100)

k0_bsw1 = []
k2l_bsw1 = []
k0_bsw2 = []
k2l_bsw2 = []
qx = []
qy = []
bety_at_mker = []
bety_at_mker_uncorrected = []
qy_uncorrected = []
for ii, tt in enumerate(t_test):
    print(f'Twiss at t = {tt*1e3:.2f} ms   ', end='\r', flush=True)
    line.vars['t_turn_s'] = tt

    line.vars['on_chicane_beta_corr'] = 1
    line.vars['on_chicane_tune_corr'] = 1
    tw = line.twiss()

    qx.append(tw.qx)
    qy.append(tw.qy)
    bety_at_mker.append(tw['bety', 'mker_match'])
    k0_bsw1.append(line['bi1.bsw1l1.1'].k0)
    k2l_bsw1.append(line['bi1.bsw1l1.1'].knl[2])
    k0_bsw2.append(line['bi1.bsw1l1.2'].k0)
    k2l_bsw2.append(line['bi1.bsw1l1.2'].knl[2])

    line.vars['on_chicane_beta_corr'] = 0
    line.vars['on_chicane_tune_corr'] = 0
    tw_uncorr = line.twiss()
    bety_at_mker_uncorrected.append(tw_uncorr['bety', 'mker_match'])
    qy_uncorrected.append(tw_uncorr.qy)

import matplotlib.pyplot as plt
plt.close('all')
plt.figure(1)
sp1 = plt.subplot(2,1,1)
plt.plot(t_test*1e3, k0_bsw1, label='k0 bsw1')
plt.plot(t_test*1e3, k0_bsw2, label='k0 bsw2')
plt.legend()
plt.subplot(2,1,2, sharex=sp1)
plt.plot(t_test*1e3, k2l_bsw1, label='k2l bsw1')
plt.plot(t_test*1e3, k2l_bsw2, label='k2l bsw2')
plt.legend()
plt.xlabel('time [ms]')

plt.figure(2)
sp1 = plt.subplot(2,1,1, sharex=sp1)
plt.plot(t_test*1e3, qy, label='qy')
plt.plot(t_test*1e3, qy_uncorrected, label='qy (uncorrected)')
plt.legend()
sp2 = plt.subplot(2,1,2, sharex=sp1)
plt.plot(t_test*1e3, bety_at_mker , label='bety at marker')
plt.plot(t_test*1e3, bety_at_mker_uncorrected, label='bety at marker (uncorrected)')
plt.legend()
plt.xlabel('time [ms]')


plt.figure(3)
sp1 = plt.subplot(2,1,1)
plt.plot(t_correct*1e3, kbrqf_corr_list, label='kbrqfcorr')
plt.plot(t_correct*1e3, kbrqd_corr_list, label='kbrqdcorr')
plt.legend()
plt.subplot(2,1,2, sharex=sp1)
plt.plot(t_correct*1e3, kbrqd3corr_list, label='kbrqd3corr')
plt.plot(t_correct*1e3, kbrqd14corr_list, label='kbrqd14corr')
plt.legend()
plt.xlabel('time [ms]')

plt.show()



import numpy as np
import math

def get_amplitude_detuning(line, nemitt_x=1.e-6, nemitt_y=1.e-6, num_turns=100, zeropad=100000):
    '''
    returns axx, axy, ayx, ayy that correspond to the first order amplitude-detuning,
    estimated through tracking.
    The relations between tune (Q) and action (J) in this case is:
    $$ Q_x = Q_{x,0} + axx * Jx + axy * Jy $$
    $$ Q_y = Q_{y,0} + ayx * Jx + ayy * Jy $$
    '''

    egeom_x = nemitt_x / line.particle_ref.gamma0
    egeom_y = nemitt_y / line.particle_ref.gamma0

    frequency = np.fft.fftfreq(zeropad)

    sigma = 2
    num_r = 50
    JJ = np.linspace(0.01, sigma**2/2., num_r)
    A_norm = np.sqrt(2*JJ).flatten()
    other_norm = 0.01

    particles = line.build_particles(x_norm=A_norm, y_norm=other_norm,
                                     nemitt_x=nemitt_x, nemitt_y=nemitt_y)

    line.track(particles, num_turns=num_turns, turn_by_turn_monitor=True)
    # line.tracker._context.synchronize()

    x = line.record_last_track.x
    y =  line.record_last_track.y

    qx = [abs(frequency[np.argmax(np.abs(
        np.fft.fft(x[ii]*np.hanning(x.shape[1]),n=zeropad)))])
        for ii in range(x.shape[0])]
    qy = [abs(frequency[np.argmax(np.abs(
        np.fft.fft(y[ii]*np.hanning(x.shape[1]), n=zeropad)))])
        for ii in range(x.shape[0])]

    axx = np.polyfit(JJ*egeom_x, qx, 1)[0]
    ayx = np.polyfit(JJ*egeom_x, qy, 1)[0]

    # switch x and y
    particles = line.build_particles(x_norm=other_norm, y_norm=A_norm,
                                     nemitt_x=nemitt_x, nemitt_y=nemitt_y)

    line.track(particles, num_turns=num_turns, turn_by_turn_monitor=True)
    # line.tracker._context.synchronize()

    x = line.record_last_track.x
    y = line.record_last_track.y

    qx = [abs(frequency[np.argmax(np.abs(np.fft.fft(x[ii]*np.hanning(x.shape[1]), n=zeropad)))]) for ii in range(x.shape[0])]
    qy = [abs(frequency[np.argmax(np.abs(np.fft.fft(y[ii]*np.hanning(x.shape[1]), n=zeropad)))]) for ii in range(x.shape[0])]

    axy = np.polyfit(JJ*egeom_y, qx, 1)[0]
    ayy = np.polyfit(JJ*egeom_y, qy, 1)[0]

    return axx, axy, ayx, ayy

class Chromaticity(object):
    def __init__(self, deltas, qx, qy, order):
        self.deltas = deltas
        self.qx = qx
        self.qy = qy
        self.order = order
        factorials = np.array([math.factorial(ii) for ii in range(order + 1)])
        self.qx_derivatives = np.polynomial.polynomial.Polynomial.fit(deltas, qx, order).convert().coef / factorials
        self.qy_derivatives = np.polynomial.polynomial.Polynomial.fit(deltas, qy, order).convert().coef / factorials


def get_nonlinear_chromaticity(line, max_delta=1.e-3, npoints=21, order=2):
    '''
    returns an object containing chromaticity information.
    It includes the derivatives of tune with respect to delta (Q', Q'', ...), estimated 
    through off-momentum twiss evaluations
    $$ Q^{(n)}_x = \frac{\partial Q}{\partial \delta} $$.
    '''
    twiss4d = line.twiss(method='4d')
    dx = twiss4d.dx[0]
    dpx = twiss4d.dpx[0]
    dy = twiss4d.dy[0]
    dpy = twiss4d.dpy[0]

    deltas = np.linspace(-max_delta, max_delta, npoints)
    qx = np.zeros(len(deltas))
    qy = np.zeros(len(deltas))

    for ii,delta in enumerate(deltas):
        part_co_guess = line.particle_ref.copy()
        part_co_guess.x = delta*dx
        part_co_guess.px = delta*dpx
        part_co_guess.y = delta*dy
        part_co_guess.py = delta*dpy
        twiss = line.twiss(method='4d', delta0=delta, co_guess=part_co_guess)
        qx[ii] = twiss.qx
        qy[ii] = twiss.qy

    chromaticity = Chromaticity(deltas, qx, qy, order)
    return chromaticity
